#include "stdafx.h"
#include "base/NativeApp.h"
#include "base/stringutil.h"
#include "file/ini_file.h"
#include "input/input_state.h"
#include "ext/libpng17/png.h"
#include "util/text/utf8.h"
#include "Common/FileUtil.h"
#include "Core/Config.h"
#include "GPU/GLES/TextureScaler.h"

class WindowsAudioBackend;

struct InputState input_state;

std::string System_GetProperty(SystemProperty prop) {
	return "";
}

int System_GetPropertyInt(SystemProperty prop) {
	return -1;
}

bool System_InputBoxGetString(const char *title, const char *defaultValue, char *outValue, size_t outLength) {
	return false;
}

bool System_InputBoxGetWString(const wchar_t *title, const std::wstring &defaultvalue, std::wstring &outvalue) {
	return false;
}

void System_SendMessage(const char *command, const char *parameter) {
}

void LaunchBrowser(const char *url) {
}

void Vibrate(int length_ms) {
}

PermissionStatus System_GetPermissionStatus(SystemPermission permission) {
	return PERMISSION_STATUS_GRANTED;
}

void System_AskForPermission(SystemPermission permission) {
}

WindowsAudioBackend *CreateAudioBackend(AudioBackendType type) {
	return nullptr;
}

void RunTests() {
}

bool IsVistaOrHigher() {
	return false;
}

namespace MainWindow {
	HWND GetHWND() {
		return 0;
	}

	void BrowseAndBoot(std::string, bool) {
	}
}

namespace W32Util {
	std::string BrowseForFolder(HWND, const char *) {
		return "";
	}
}

u8 *LoadPNG(const std::string &filename, int &w, int &h) {
	png_image png = {};
	png.version = PNG_IMAGE_VERSION;

	FILE *fp = File::OpenCFile(filename, "rb");
	if (!png_image_begin_read_from_stdio(&png, fp)) {
		fprintf(stderr, "Could not read png: %s\n", png.message);
		return nullptr;
	}

	u8 *out = new u8[png.width * png.height * 4];

	png.format = PNG_FORMAT_RGBA;
	if (!png_image_finish_read(&png, nullptr, out, png.width * 4, nullptr)) {
		fprintf(stderr, "Could not read png: %s\n", png.message);
		delete [] out;
		return nullptr;
	}

	fclose(fp);
	png_image_free(&png);

	w = png.width;
	h = png.height;
	return out;
}

static bool WriteTextureToPNG(png_imagep image, const std::string &filename, int convert_to_8bit, const void *buffer, png_int_32 row_stride, const void *colormap) {
	if (!File::Copy(filename, filename + "~")) {
		fprintf(stderr, "Unable to backup old file.\n");
		return false;
	}

	FILE *fp = File::OpenCFile(filename, "wb");
	if (!fp) {
		fprintf(stderr, "Unable to open texture file for writing.\n");
		return false;
	}

	if (png_image_write_to_stdio(image, fp, convert_to_8bit, buffer, row_stride, colormap)) {
		if (fclose(fp) != 0) {
			fprintf(stderr, "Texture file write failed.\n");
			return false;
		}
		return true;
	} else {
		fprintf(stderr, "Texture PNG encode failed.\n");
		fclose(fp);
		remove(filename.c_str());
		return false;
	}
}

bool SavePNG(const std::string &filename, u32 *data, int w, int h, int stride) {
	png_image png;
	memset(&png, 0, sizeof(png));
	png.version = PNG_IMAGE_VERSION;
	png.format = PNG_FORMAT_RGBA;
	png.width = w;
	png.height = h;
	bool success = WriteTextureToPNG(&png, filename, 0, data, stride * 4, nullptr);
	png_image_free(&png);

	if (png.warning_or_error >= 2) {
		fprintf(stderr, "Could not write png: %s\n", png.message);
		return false;
	}

	return true;
}

int ParseScalingType(const std::string &t) {
	if (strcasecmp(t.c_str(), "xbrz") == 0) {
		return TextureScaler::XBRZ;
	} else if (strcasecmp(t.c_str(), "hybrid") == 0) {
		return TextureScaler::HYBRID;
	} else if (strcasecmp(t.c_str(), "bicubic") == 0) {
		return TextureScaler::BICUBIC;
	} else if (strcasecmp(t.c_str(), "hybrid_bicubic") == 0) {
		return TextureScaler::HYBRID_BICUBIC;
	}
	return -1;
}

int ParseScaleFactor(const std::string &f) {
	int value = -1;
	if (!TryParse(f, &value)) {
		return -1;
	}
	if (value < 0 || value > 5) {
		return -1;
	}
	return value;
}

int wmain(int argc, wchar_t *argv[]) {
	std::vector<std::string> filenames;
	g_Config.iTexScalingType = TextureScaler::HYBRID;
	g_Config.bTexDeposterize = true;
	int scaleFactor = 3;
	int forceWidth = -1;
	int forceHeight = -1;

	for (int i = 1; i < argc; ++i) {
		if (argv[i][0] == '-') {
			if (wcscmp(argv[i], L"--type") == 0) {
				g_Config.iTexScalingType = ParseScalingType(ConvertWStringToUTF8(argv[++i]));
				if (g_Config.iTexScalingType == -1) {
					fprintf(stderr, "Unexpected --type value: %S (xbrz, hybrid, bicubic, hybrid_bicubic)\n", argv[i]);
					return 1;
				}
				continue;
			} else if (wcsncmp(argv[i], L"--type=", 7) == 0) {
				g_Config.iTexScalingType = ParseScalingType(ConvertWStringToUTF8(argv[i] + 7));
				if (g_Config.iTexScalingType == -1) {
					fprintf(stderr, "Unexpected --type value: %S (xbrz, hybrid, bicubic, hybrid_bicubic)\n", argv[i] + 7);
					return 1;
				}
				continue;
			} else if (wcscmp(argv[i], L"--no-deposterize") == 0) {
				g_Config.bTexDeposterize = false;
				continue;
			} else if (wcscmp(argv[i], L"--scale") == 0) {
				scaleFactor = ParseScaleFactor(ConvertWStringToUTF8(argv[++i]));
				if (scaleFactor == -1) {
					fprintf(stderr, "Unexpected --scaling value: %S\n", argv[i]);
					return 1;
				}
				continue;
			} else if (wcsncmp(argv[i], L"--scale=", 8) == 0) {
				scaleFactor = ParseScaleFactor(ConvertWStringToUTF8(argv[i] + 8));
				if (scaleFactor == -1) {
					fprintf(stderr, "Unexpected --scaling value: %S\n", argv[i] + 8);
					return 1;
				}
				continue;
			} else if (wcscmp(argv[i], L"--width") == 0) {
				if (!TryParse(ConvertWStringToUTF8(argv[++i]), &forceWidth)) {
					fprintf(stderr, "Unexpected --width value: %S\n", argv[i]);
					return 1;
				}
				continue;
			} else if (wcsncmp(argv[i], L"--width=", 8) == 0) {
				if (!TryParse(ConvertWStringToUTF8(argv[i] + 8), &forceWidth)) {
					fprintf(stderr, "Unexpected --width value: %S\n", argv[i] + 8);
					return 1;
				}
				continue;
			} else if (wcscmp(argv[i], L"--height") == 0) {
				if (!TryParse(ConvertWStringToUTF8(argv[++i]), &forceHeight)) {
					fprintf(stderr, "Unexpected --height value: %S\n", argv[i]);
					return 1;
				}
				continue;
			} else if (wcsncmp(argv[i], L"--height=", 9) == 0) {
				if (!TryParse(ConvertWStringToUTF8(argv[i] + 9), &forceHeight)) {
					fprintf(stderr, "Unexpected --height value: %S\n", argv[i] + 8);
					return 1;
				}
				continue;
			}
		}

		filenames.push_back(ConvertWStringToUTF8(argv[i]));
	}
	if (filenames.empty()) {
		fprintf(stderr, "Usage: %S [--type hybrid] [--no-deposterize] [--scale 3] [--width 30] [--height 60] filename.png...\n", argv[0]);
		return 1;
	}

	for (const std::string &filename : filenames) {
		int w, h;
		u8 *out = LoadPNG(filename, w, h);
		if (!out) {
			return 1;
		}

		u32 *newOut = (u32 *)out;
		TextureScalerGL scaler;
		u32 dstFmt = 0x1401;
		scaler.Scale(newOut, dstFmt, w, h, scaleFactor);

		int origW = w;
		if (forceWidth != -1) {
			if (forceWidth * scaleFactor > w) {
				fprintf(stderr, "Warning: --width greater than png, ignoring.\n");
			} else {
				w = forceWidth * scaleFactor;
			}
		}
		if (forceHeight != -1) {
			if (forceHeight * scaleFactor > h) {
				fprintf(stderr, "Warning: --height greater than png, ignoring.\n");
			} else {
				h = forceHeight * scaleFactor;
			}
		}

		if (!SavePNG(filename, newOut, w, h, origW)) {
			return 1;
		}

		delete [] out;
	}

    return 0;
}
